#include "ArtNetController.h"

//********************************************************
//**Art-Net™ Designed by and Copyright Artistic Licence.**
//********************************************************

ArtNetController::ArtNetController(){}

ArtNetController::ArtNetController (DmxController* dmxController, PresetRepository* presetRepository)
{
    this->startUdp();
    this->dmxController = dmxController;
    this->presetRepository = presetRepository;

    for(int i = 0; i < 4;i++)
    {
        this->controllerIpAddress[i] = 0x00;
    }
}

int ArtNetController::startUdp()
{
    this->mac[0] = MAC_ADDRESS_PART_01;
    this->mac[1] = MAC_ADDRESS_PART_02;
    this->mac[2] = MAC_ADDRESS_PART_03;
    this->mac[3] = MAC_ADDRESS_PART_04;
    this->mac[4] = MAC_ADDRESS_PART_05;
    this->mac[5] = MAC_ADDRESS_PART_06;

    if (Ethernet.begin(this->mac) == 0)
    {
        Serial.println("Failed to configure Ethernet using DHCP");
        if (Ethernet.hardwareStatus() == EthernetNoHardware)
        {
            Serial.println("Ethernet shield was not found.  Sorry, can't run without hardware. :(");
        }
        else if (Ethernet.linkStatus() == LinkOFF)
        {
            Serial.println("Ethernet cable is not connected.");
        }
    }

    this->nodeIpAddress[0] = Ethernet.localIP()[0];
    this->nodeIpAddress[1] = Ethernet.localIP()[1];
    this->nodeIpAddress[2] = Ethernet.localIP()[2];
    this->nodeIpAddress[3] = Ethernet.localIP()[3];

    this->udp.begin(UDP_PORT_ADRESS);
}

bool ArtNetController::checkArtNetHeader(uint8_t udpData[])
{
    bool headerFound = true;
    uint8_t artNetHeader[8] = {'A','r','t','-','N','e','t',0x00};
    for(int i = 0; i < 8; i++)
    {
        if(udpData[i] != artNetHeader[i])
        {
            headerFound = false;
        }
    }

    return headerFound;
}

uint8_t *ArtNetController::getPresetStatus()
{
    uint8_t *presetsStatus = new uint8_t[NUMBER_OF_PRESETS];
    for(int i = 0; i < NUMBER_OF_PRESETS; i++)
    {
        presetsStatus[i] = this->presetRepository->getPresetState(i);
    }

    return presetsStatus;
}

void ArtNetController::sendPresetActivated(int presetId)
{
    String commandString = String(ART_COM_PA) + "={\"presetId\":" + String(presetId) + "}";
    uint16_t length = commandString.length();
    char command[length];
    commandString.toCharArray(command, length+1);
    this->sendArtCommand(command, length);
}

void ArtNetController::sendPresetDeactivated(int presetId)
{
    String commandString = String(ART_COM_PD) + "={\"presetId\":" + String(presetId) + "}";
    uint16_t length = commandString.length();
    char command[length];
    commandString.toCharArray(command, length+1);
    this->sendArtCommand(command, length);
}

void ArtNetController::listenToArtNet()
{
    
    int packetSize = this->udp.parsePacket();
    if(packetSize >= 14)
    {
        // read the packet into packetBufferRead
        uint8_t *packetBufferRead =  new uint8_t[packetSize];
        this->udp.read(packetBufferRead,packetSize);

        if(this->checkArtNetHeader(packetBufferRead))
        {
            uint16_t opCode = (((uint16_t)packetBufferRead[9] << 8) + packetBufferRead[8]);
            switch(opCode)
            {
                case OP_POLL:
                    // save ip address from controller
                    for(int i = 0; i < 4;i++)
                    {
                        this->controllerIpAddress[i] = this->udp.remoteIP()[i];
                    }
                    this->sendArtPollReply();
                    break;

                case OP_COMMAND:
                    this->processArtCommand(packetBufferRead, packetSize);
                    break;

                case OP_DATA_REQUEST:
                    // save ip address from controller
                    for(int i = 0; i < 4;i++)
                    {
                        this->controllerIpAddress[i] = this->udp.remoteIP()[i];
                    }
                    this->processArtDataRequest(packetBufferRead, packetSize);
                    break;
                
                case OP_DMX:
                    this->processArtDmx(packetBufferRead, packetSize);
            }
        }
    }
}

ArtCommand ArtNetController::parseToArtCommand(uint8_t updData[], int length)
{
    ArtCommand artCommand;

    if(length >= 16)
    {
        // protocoll version
        artCommand.ProtVerHi = updData[10];
        artCommand.ProtVerLo = updData[11];

        // esta manifacture
        artCommand.EstaManHi = updData[12];
        artCommand.EstaManLo = updData[13];

        // lengths
        artCommand.Length = (((uint16_t)updData[14] << 8) + updData[15]);

        if(artCommand.Length > 0)
        {
            artCommand.Data = new char[artCommand.Length + 1];
            for(int i = 0; i < artCommand.Length; i++)
            {
                artCommand.Data[i] = (char)updData[16 + i];
            }
            artCommand.Data[artCommand.Length] = '\0';
        }
    }

    return artCommand;    
}

ArtDataRequest ArtNetController::parseToArtDataRequest(uint8_t updData[], int length)
{
    ArtDataRequest artDataRequest;
    if(length >= 18)
    {
        // protocoll version
        artDataRequest.ProtVerHi = updData[10];
        artDataRequest.ProtVerLo = updData[11];

        // esta manifacture
        artDataRequest.EstaManHi = updData[12];
        artDataRequest.EstaManLo = updData[13];

        // oem
        artDataRequest.OemHi = updData[14];
        artDataRequest.OemLo = updData[15];

        // request code
        artDataRequest.RequestCode = (((uint16_t)updData[16] << 8) + updData[17]);
    }

    return artDataRequest;
}

ArtDmx ArtNetController::parseToArtDmx(uint8_t updData[], int length)
{
    ArtDmx artDmx;
    if(length >= 18)
    {
        // protocoll version
        artDmx.ProtVerHi = updData[10];
        artDmx.ProtVerLo = updData[11];
        
        // sequence
        artDmx.Sequence = updData[12];

        // pysical input port
        artDmx.Physical = updData[13];
        
        // art-net address
        artDmx.SubUni = updData[14];
        artDmx.Net = updData[15];

        artDmx.Length = (((uint16_t)updData[16] << 8) + updData[17]);

        if(artDmx.Length > 0)
        {
            artDmx.Data = new uint8_t[artDmx.Length];
            for(int i = 0; i < artDmx.Length; i++)
            {
                artDmx.Data[i] = updData[18+i];
            }
        }
    }

    return artDmx;
}

void ArtNetController::parseFromArtPollReply(ArtPollReply artPollReply, uint8_t* udpSendData, int length)
{
    if(length >= 207)
    {
        // id
        for(int i = 0; i<8; i++)
        {
            udpSendData[i] = artPollReply.Id[i];
        }
        
        // op code
        udpSendData[8] = artPollReply.OpCode;
        udpSendData[9] = artPollReply.OpCode >> 8;
        
        // node ip
        for(int i= 0; i < 4; i++)
        {
            udpSendData[10+i] = artPollReply.IpAddress[i];
        }

        // port
        udpSendData[14] = artPollReply.Port;
        udpSendData[15] = artPollReply.Port>> 8;
        
        // version info
        udpSendData[16] = artPollReply.VersInfoH;
        udpSendData[17] = artPollReply.VersInfoL;
        
        // net address
        udpSendData[18] = artPollReply.NetSwitch;
        
        // sub net address
        udpSendData[19] = artPollReply.SubSwitch;
        
        // oem
        udpSendData[20] = artPollReply.OemHi;
        udpSendData[21] = artPollReply.OemLo;
        
        // ubea version
        udpSendData[22] = artPollReply.UbeaVersion;
        
        // status
        udpSendData[23] = artPollReply.Status1;
        
        // esta
        udpSendData[24] = artPollReply.EstaManLo;
        udpSendData[25] = artPollReply.EstaManHi;
        
        // port name
        for (int i = 0; i < 18; i++)
        {
            udpSendData[26+i] = artPollReply.PortName[i];
        }

        // long name
        for (int i = 0; i < 64; i++)
        {
            udpSendData[44+i] = artPollReply.LongName[i];
        }

        // node report
        for (int i = 0; i < 64; i++)
        {
            udpSendData[108+i] = artPollReply.NodeReport[i];
        }

        // number of ports
        udpSendData[172] = artPollReply.NumPortsHi;
        udpSendData[173] = artPollReply.NumPortsLo;

        // port type
        for (int i = 0; i < 4; i++)
        {
            udpSendData[174+i] = artPollReply.PortTypes[i];
        }

        // good inputs
        for (int i = 0; i < 4; i++)
        {
            udpSendData[178+i] = artPollReply.GoodInput[i];
        }

        // good output A
        for (int i = 0; i < 4; i++)
        {
            udpSendData[182+i] = artPollReply.GoodOutputA[i];
        }

        // sw in
        for (int i = 0; i < 4; i++)
        {
            udpSendData[186+i] = artPollReply.SwIn[i];
        }

        // sw out
        for (int i = 0; i < 4; i++)
        {
            udpSendData[190+i] = artPollReply.SwOut[i];
        }

        // prio acn
        udpSendData[194] = artPollReply.AcnPriority;

        // sw macro
        udpSendData[195] = artPollReply.SwMacro;
        
        // sw remote
        udpSendData[196] = artPollReply.SwRemote;

        // spare
        for (int i = 0; i < 3; i++)
        {
            udpSendData[197+i] = 0x00;
        }

        // style
        udpSendData[200] = artPollReply.Style;

        // mac address
        for (int i = 0; i < 6; i++)
        {
            udpSendData[201+i] = artPollReply.Mac[i];
        }
    }
    if(length >= 208)
    {
        // bind ip
        for (int i = 0; i < 4; i++)
        {
            udpSendData[207+i] = artPollReply.BindIp[i];
        }
    }
    if(length >= 212)
    {
        // bind index
        udpSendData[211] = artPollReply.BindIndex;
    }
    if(length >= 213)
    {
        // status 2
        udpSendData[212] = artPollReply.Status2;
    }
    if(length >= 214)
    {
        // good output b
        for (int i = 0; i < 4; i++)
        {
            udpSendData[213+i] = artPollReply.GoodOutputB[i];
        }
    }
    if(length >= 218)
    {
        // status 3
        udpSendData[217] = artPollReply.Status3;
    }
    if(length >= 219)
    {
        // default responder uid
        for (int i = 0; i < 6; i++)
        {
            udpSendData[218+i] = artPollReply.DefaulRespUID[i];
        }
    }
    if(length >= 225)
    {
        // user data
        udpSendData[224] = artPollReply.UserHi;
        udpSendData[225] = artPollReply.UserLo;
    }
    if(length >= 227)
    {
        // user data
        udpSendData[226] = artPollReply.RefreshRateHi;
        udpSendData[227] = artPollReply.RefreshRateLo;

        // filler
        for (int i = 0; i < 8; i++)
        {
            udpSendData[228+i] = 0x00;
        }

    }
}

void ArtNetController::parseFromArtCommand(ArtCommand artCommand, uint8_t* udpSendData, int length)
{
    if(length > 16)
    {
        // id
        for(int i = 0; i<8; i++)
        {
            udpSendData[i] = artCommand.Id[i];
        }
        
        // op code
        udpSendData[8] = artCommand.OpCode;
        udpSendData[9] = artCommand.OpCode >> 8;

        // protocol version
        udpSendData[10] = artCommand.ProtVerHi;
        udpSendData[11] = artCommand.ProtVerLo;

        // esta manifactur
        udpSendData[12] = artCommand.EstaManHi;
        udpSendData[13] = artCommand.EstaManLo;

        // length
        udpSendData[14] = artCommand.Length >> 8;
        udpSendData[15] = artCommand.Length;

        for(int i=0; i < artCommand.Length; i++)
        {
            udpSendData[16+i] = artCommand.Data[i];
        }
    }
}

void ArtNetController::parseFromArtDataReply(ArtDataReply artDataReply, uint8_t* udpSendData, int length)
{
    if(length >= 20)
    {
        // id
        for(int i = 0; i<8; i++)
        {
            udpSendData[i] = artDataReply.Id[i];
        }
        
        // op code
        udpSendData[8] = artDataReply.OpCode;
        udpSendData[9] = artDataReply.OpCode >> 8;

        // protocol version
        udpSendData[10] = artDataReply.ProtVerHi;
        udpSendData[11] = artDataReply.ProtVerLo;

        // esta manifactur
        udpSendData[12] = artDataReply.EstaManHi;
        udpSendData[13] = artDataReply.EstaManLo;

        // oem
        udpSendData[14] = artDataReply.OemHi;
        udpSendData[15] = artDataReply.OemLo;

        // request code
        udpSendData[16] = artDataReply.RequestCode >> 8;
        udpSendData[17] = artDataReply.RequestCode;

        // peyload lengths
        udpSendData[18] = artDataReply.PayLength >> 8;
        udpSendData[19] = artDataReply.PayLength;

        for(int i=0; i < artDataReply.PayLength; i++)
        {
            udpSendData[20+i] = artDataReply.PayLoad[i];
        }
    }
}

void ArtNetController::sendArtPollReply()
{
    int packageSize = 236;

    ArtPollReply artPollReply;
    // set controller ip
    for(int i = 0; i < 4; i++)
    {
        artPollReply.IpAddress[i] = this->nodeIpAddress[i];
    }

    char* longName = "KIWI (is a) Illumination Weeny Interface";

    // set long name
    bool endOfLine = false;
    for(int i = 0; i < 64; i++)
    {
        if(longName[i] != '\0' && !endOfLine)
        {
            artPollReply.LongName[i] = longName[i];
        }
        else
        {
            endOfLine = true;
            artPollReply.LongName[i] = 0x00;
        }
    }

    // set node report
    for(int i = 0; i < 64; i++)
    {
        artPollReply.NodeReport[i] = 0x00;
    }

    // set mac address
    for(int i = 0; i < 6; i++)
    {
        artPollReply.Mac[i] = this->mac[i];
    }
    
    uint8_t updSendData[packageSize];
    this->parseFromArtPollReply(artPollReply, updSendData, packageSize);


    IPAddress remoteIp = IPAddress(this->controllerIpAddress);
    this->udp.beginPacket(remoteIp, UDP_PORT_ADRESS);
    this->udp.write(updSendData, packageSize);
    this->udp.endPacket();
}

void ArtNetController::sendArtCommand(char *message, uint16_t length)
{
    ArtCommand artCommand;
    artCommand.Length = length;
    artCommand.Data = message;
    int packageSize = 16 + length;

    if(this->controllerIpAddress[0] != 0x00)
    {
        uint8_t updSendData[packageSize];
        this->parseFromArtCommand(artCommand, updSendData, packageSize);

        IPAddress remoteIp = IPAddress(this->controllerIpAddress);
        this->udp.beginPacket(remoteIp, UDP_PORT_ADRESS);
        this->udp.write(updSendData, packageSize);
        this->udp.endPacket(); 
    }    
}

void ArtNetController::sendArtDataReply(uint16_t dataRequstCode)
{
    ArtDataReply artDataReply;
    artDataReply.RequestCode = dataRequstCode;
    int packageSize = 20;
    uint8_t productUrl[PRODUCT_URL_LENGTHS];
    for(int i = 0; i < PRODUCT_URL_LENGTHS; i++)
    {
        productUrl[i] = String(PRODUCT_URL).charAt(i);
    }

    switch(dataRequstCode)
    {
        case DR_POLL:
            artDataReply.PayLength = 0x0000;
            artDataReply.PayLoad = nullptr;
            break;
        case DR_URL_PRODUCT:
            artDataReply.PayLength = PRODUCT_URL_LENGTHS;
            artDataReply.PayLoad = productUrl;
            packageSize += PRODUCT_URL_LENGTHS;
            break;
        case DR_URL_USER_GUIDE:
            artDataReply.PayLength = PRODUCT_URL_LENGTHS;
            artDataReply.PayLoad = productUrl;
            packageSize += PRODUCT_URL_LENGTHS;
            break;
        case DR_URL_SUPPORT:
            artDataReply.PayLength = PRODUCT_URL_LENGTHS;
            artDataReply.PayLoad = productUrl;
            packageSize += PRODUCT_URL_LENGTHS;
            break;
        case DR_URL_PERS_UDR:
            artDataReply.PayLength = 0x0000;
            artDataReply.PayLoad = nullptr;
            break;
            break;
        case DR_URL_PERS_GDTF:
            artDataReply.PayLength = 0x0000;
            artDataReply.PayLoad = nullptr;
            break;
            break;
        case DR_PRESET_STATUS:
            artDataReply.PayLength = (uint16_t)NUMBER_OF_PRESETS;
            artDataReply.PayLoad = getPresetStatus();
            packageSize += NUMBER_OF_PRESETS;
            break;
    }

    uint8_t updSendData[packageSize];
    this->parseFromArtDataReply(artDataReply, updSendData, packageSize);

    IPAddress remoteIp = IPAddress(this->controllerIpAddress);
    this->udp.beginPacket(remoteIp, UDP_PORT_ADRESS);
    this->udp.write(updSendData, packageSize);
    this->udp.endPacket();
}

void ArtNetController::processArtCommand(uint8_t updData[], int length)
{
    ArtCommand artCommand = this->parseToArtCommand(updData, length);
    
    char *commandString = strtok(artCommand.Data, "&");
    while (commandString != NULL) {
        String command = String(commandString).substring(0, String(commandString).indexOf('='));
        int commandValuelength = (String(commandString).substring(String(commandString).indexOf('=')+1)).length() + 1;
        char commandValue[commandValuelength];
        String(commandString).substring(String(commandString).indexOf('=')+1).toCharArray(commandValue, commandValuelength);
        deserializeJson(doc, commandValue, commandValuelength);

        if(command.equals(ART_COM_AP))
        {   
            this->presetRepository->activatePreset(doc[activatePresetStructure().presetId]);
        }
        else if(command.equals(ART_COM_DP))
        {
            this->presetRepository->deactivatePreset(doc[deactivatePresetStructure().presetId]);
        }
        else if(command.equals(ART_COM_CP))
        {
            this->presetRepository->clearPreset(doc[clearPresetStructure().presetId]);
        }
        else if(command.equals(ART_COM_SP))
        {
            this->presetRepository->setPresetGroupId(doc[setPresetStructure().presetId], doc[setPresetStructure().groupId]);
            this->presetRepository->setPresetChannelValue(doc[setPresetStructure().presetId], doc[setPresetStructure().channel], doc[setPresetStructure().value]);
        }
        else if(command.equals(ART_COM_SC))
        {
            this->dmxController->setChannelStatus(doc[setChannelStateStructure().isActive], doc[setChannelStateStructure().channel]);
            this->presetRepository->updatePresetState();
        }
        else if(command.equals(ART_COM_AG))
        {
            this->presetRepository->activateGroup(doc[activateGroupStructure().groupId]);
        }
        else if(command.equals(ART_COM_DG))
        {
            this->presetRepository->deactivateGroup(doc[deactivateGroupStructure().groupId]);
        }

        commandString = strtok(NULL, "&");
    }
}

void ArtNetController::processArtDataRequest(uint8_t updData[], int length)
{
    ArtDataRequest artDataRequest = this->parseToArtDataRequest(updData, length);
    this->sendArtDataReply(artDataRequest.RequestCode);
}

void ArtNetController::processArtDmx(uint8_t updData[], int length)
{
    ArtDmx artDmx = this->parseToArtDmx(updData, length);
    if(artDmx.SubUni == 0x00 && artDmx.Net == 0x00)
    {
        for(int i = 0; i<artDmx.Length; i++)
        {
            this->dmxController->setChannelValue(i+1, artDmx.Data[i]);
        }
    }
}