#include <ArduinoSTL.h>
#include <DmxSimple.h>
#include "Preset.h"
#include <SET.h>

#define MAX_DMX_CHANNEL 100
#define MAX_LAMP_CHANNEL 5
#define PIN_DMX_OUT 26
#define NUMBER_OF_PRESETS 6
#define BUFFER_SIZE 5

int lampPins[] = {33,31,35,37,32,30};
int buttonPins[] = {29,27,25,23,22,24};
char serialBuffer[BUFFER_SIZE];
std::vector<Preset> presets;
std::vector<byte> dmxOut(MAX_DMX_CHANNEL);


// dmx functions
void sendDmx() {
  for(int i = 0; i < MAX_DMX_CHANNEL; i++){
    DmxSimple.write(i+1, dmxOut[i]);
  }
}

void updateDmx(std::vector<byte>& values, int reset) {
  for(int i = 0; i < MAX_DMX_CHANNEL; i++){
    if(values[i]) {
      if(!reset) {
        dmxOut[i] = values[i];
      } else {
        dmxOut[i] = 0;
      }
    }
  }
}

// buttons functions
void checkButtons() {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++) {
    if(digitalRead(buttonPins[i]) == LOW) {
      if(presets[i].getPresetStatus() == Preset::OFF) {
        activatePreset(i);
      } else {
        deactivateGroup(presets[i].getGroupId());
      }
      while(digitalRead(buttonPins[i]) == LOW) {}
    }
  }
}

// preset functions
void loadPresets() {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++){
    presets[i].loadDmxArray();
  }
}

void activatePreset(int presetNr) {
  deactivateOtherGroupPresets(presetNr);
  presets[presetNr].setPresetStatus(Preset::ACTIVE);
  std::vector<byte> values = presets[presetNr].getDmxArray();
  updateDmx(values, 0);
  char buff[] = {'P','A',presetNr,0,0};
  Serial.write(buff, BUFFER_SIZE);
}

void deactivatePreset(int presetNr) {
  presets[presetNr].setPresetStatus(Preset::OFF);
  std::vector<byte> values = presets[presetNr].getDmxArray();
  updateDmx(values, 1);
  char buff[] = {'P','D',presetNr,0,0};
  Serial.write(buff, BUFFER_SIZE);
}

void clearPreset(int presetNr) {
  deactivatePreset(presetNr);
  std::vector<byte> dmxData(MAX_DMX_CHANNEL);
  presets[presetNr].setDmxArray(dmxData);
}

void setPreset(int presetNr, int channelNr, int value){
  std::vector<byte> dmxData(MAX_DMX_CHANNEL);
  dmxData = presets[presetNr].getDmxArray();
  dmxData[channelNr - 1] = value;
  presets[presetNr].setDmxArray(dmxData);
}

// group functions
void activateGroup(byte groupId) {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++) {
    if(presets[i].getPresetStatus() == Preset::OFF && presets[i].getGroupId() == groupId) {
      presets[i].setPresetStatus(Preset::ACTIVE);
    }
  }
}

void deactivateGroup(byte groupId) {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++) {
    if(presets[i].getPresetStatus() == Preset::ACTIVE && presets[i].getGroupId() == groupId) {
      deactivatePreset(i);
    }
  }
}

void deactivateOtherGroupPresets(int presetId) {
  byte groupId = presets[presetId].getGroupId();
  for(int i = 0; i < NUMBER_OF_PRESETS; i++) {
    if(i != presetId && presets[i].getPresetStatus() == Preset::ACTIVE && presets[i].getGroupId() == groupId) {
      deactivatePreset(i);
    }
  }
}

// communication functions
void serialEvent() {
  if(Serial.available()) {
    Serial.readBytes(serialBuffer, BUFFER_SIZE);

    String command = "";
    command += serialBuffer[0];
    command += serialBuffer[1];
    int presetNr = serialBuffer[2];
    int channelNr = serialBuffer[3];
    byte value = (byte)serialBuffer[4];
    
    if(command == "AP") {
      activatePreset(presetNr);
    } else if (command == "DP") {
      deactivatePreset(presetNr);
    } else if (command == "CP") {
      clearPreset(presetNr);
    } else if (command == "SP") {
      setPreset(presetNr, channelNr, value);
    } else if (command == "SC") {
      dmxOut[channelNr-1] = value;
    } else if (command == "SG") {
      presets[presetNr].setGroupId((byte)channelNr);
    } else if (command == "AG") {
      activateGroup(channelNr);
    } else if (command == "DG") {
      deactivateGroup(channelNr);
    } else if (command == "RS") {
      sendStatus();
    }
  }
}

void sendStatus() {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++) {
    if(presets[i].getPresetStatus() == Preset::ACTIVE) {
      Serial.println("PA00" + String(i) + "000000");
    }
  }
}

// setup functions
void setupButtons() {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++){
    pinMode(buttonPins[i], INPUT_PULLUP);
  }
}

void setupLamps() {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++){
    pinMode(lampPins[i], OUTPUT);
  }
}

void setupPresets() {
  for(int i = 0; i < NUMBER_OF_PRESETS; i++){
    Preset* presetPtr = new Preset(lampPins[i], i*(MAX_DMX_CHANNEL+2), MAX_DMX_CHANNEL);
    presets.push_back(*presetPtr);
    presets[i].loadDmxArray();
    presets[i].loadGroupId();
    presets[i].loadPresetStatus();
    if(presets[i].getPresetStatus() == Preset::ACTIVE) {
      activatePreset(i);
    }
  }

  
}

void setupData() {
  std::vector<byte> scene1(MAX_DMX_CHANNEL);
  scene1[0] = 255;
  scene1[1] = 255;
  scene1[2] = 255;
  scene1[3] = 255;
  scene1[4] = 255;
  scene1[5] = 255;
  scene1[6] = 255;
  scene1[7] = 255;
  scene1[8] = 255;
  scene1[9] = 255;
  scene1[10] = 255;
  scene1[11] = 255;
  presets[0].setDmxArray(scene1);
  presets[0].setGroupId(1);

  std::vector<byte> scene2(MAX_DMX_CHANNEL);
  scene2[0] = 255;
  scene2[3] = 255;
  presets[1].setDmxArray(scene2);
  presets[1].setGroupId(2);

  std::vector<byte> scene3(MAX_DMX_CHANNEL);
  scene3[1] = 255;
  scene3[4] = 255;
  presets[2].setDmxArray(scene3);
  presets[2].setGroupId(2);

  std::vector<byte> scene4(MAX_DMX_CHANNEL);
  scene4[8] = 255;
  scene4[11] = 255;
  presets[3].setDmxArray(scene4);
  presets[3].setGroupId(3);

  std::vector<byte> scene5(MAX_DMX_CHANNEL);
  scene5[6] = 255;
  scene5[8] = 255;
  scene5[9] = 255;
  scene5[11] = 255;
  presets[4].setDmxArray(scene5);
  presets[4].setGroupId(3);

  std::vector<byte> scene6(MAX_DMX_CHANNEL);
  scene6[12] = 255;
  presets[5].setDmxArray(scene6);
  
}

void setup() {
  Serial.begin(9600);
  
  setupButtons();
  setupLamps();
  setupPresets();
  //setupData();
  
  // setup DMXSimple
  DmxSimple.usePin(PIN_DMX_OUT);
  DmxSimple.maxChannel(MAX_DMX_CHANNEL);
}

// main programm
void loop() {
  // send dmx data
  sendDmx();

  // check buttons
  checkButtons();
}
